#include <stdlib.h> /* for malloc and free */ 
#include <string.h> /* for strlen, strcpy, and strcmp */ 

#include "ModuleLog.h"

#define HASH unsigned long

struct hash_entry
{
  struct hash_entry *pChain;
  HASH hash;
  char *key;
  char *value;
};

struct hash_table
{
  unsigned nTableSize, nEntries;
  struct hash_entry *Table[1];
};

/***********************************************************
* int HashTable_KeysEqual()                                *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *left;                                   *
*    - const char *right;                                  *
*                         *  *  *  *                       *
* Compares two hash keys.  Currently they are both         *
* converted to lowercase (and some other conversions), so  *
* a simple strcmp is enough.                               *
***********************************************************/
static int HashTable_KeysEqual(const char *left, const char *right)
{
  REQUIRE(READABLE(left));
  REQUIRE(READABLE(right));

  /* stricmp won't work right with MBCS */ 
  /* return CSTR_EQUAL == CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE|NORM_IGNOREKANATYPE|NORM_IGNOREWIDTH, left, -1, right, -1); */ 

  /* all keys are passed through GetHashedToken, which forces them to lower case */ 
  /* (in lsxcommand at least), so a case sensitive compare will work just fine.  */ 
  return strcmp(left, right) == 0;
}

/***********************************************************
* struct hash_table *HashTable_Alloc()                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - unsigned nSize;                                     *
*      The number of entries to initially allocate.        *
*                         *  *  *  *                       *
* Allocates memory for a hash table.                       *
***********************************************************/
static struct hash_table *HashTable_Alloc(unsigned nSize)
{
  REQUIRE(nSize > 0);
  REQUIRE((nSize & (nSize-1)) == 0);

  return malloc(sizeof(struct hash_table) + (nSize-1)*sizeof(struct hash_entry *));
}
  
/***********************************************************
* struct hash_table *HashTable_Create()                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - unsigned nSize;                                     *
*      The number of entries to initially allocate.        *
*                         *  *  *  *                       *
* Allocates and initializes a hash table.                  *
***********************************************************/
struct hash_table *HashTable_Create(unsigned nSize)
{
  struct hash_table *pTable = HashTable_Alloc(nSize);

  if (pTable)
  {
    pTable->nTableSize = nSize;
    pTable->nEntries = 0;
    do
      pTable->Table[--nSize] = NULL;
    while (nSize);
  }

  return pTable;
}

/***********************************************************
* struct hash_entry **HashTable_InternalLookup()           *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct hash_table *pTable;                          *
*    - HASH hash;                                          *
*    - const char *key;                                    *
*                         *  *  *  *                       *
* Looks up a key in the hash table and returns a pointer   *
* to the entry if it exists or to where it would go if it  *
* were to be inserted.                                     *
***********************************************************/
static struct hash_entry **HashTable_InternalLookup(
    struct hash_table *pTable,
    HASH hash,
    const char *key)
{
  unsigned mask;
  struct hash_entry **pEntry;

  REQUIRE(READABLE(pTable));
  REQUIRE(READABLEN(pTable->Table, pTable->nTableSize));

  mask = pTable->nTableSize - 1;
  pEntry = &pTable->Table[hash & mask];

  ASSERT(READABLE(pEntry));

  while (*pEntry)
  {
    REQUIRE(READABLE(*pEntry));

    if ((*pEntry)->hash == hash
        && (*pEntry)->key != NULL
        && HashTable_KeysEqual((*pEntry)->key, key))
      break;
    pEntry = &(*pEntry)->pChain;
  }

  ENSURE(READABLE(pEntry));

  return pEntry;
}

/***********************************************************
* struct hash_entry *HashTable_Lookup()                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct hash_table *pTable;                          *
*    - HASH hash;                                          *
*    - const char *key;                                    *
*                         *  *  *  *                       *
* Looks up a hash table entry and returns a pointer to the *
* entry or NULL.                                           *
***********************************************************/
static struct hash_entry *HashTable_Lookup(
    struct hash_table *pTable,
    HASH hash,
    const char *key)
{
  return *HashTable_InternalLookup(pTable, hash, key);
}

/***********************************************************
* struct hash_entry *HashTable_Insert()                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct hash_table *pTable;                          *
*    - HASH hash;                                          *
*    - const char *key;                                    *
*    - const char *value;                                  *
*                         *  *  *  *                       *
* Inserts a new entry in the hash table or replaces an     *
* existing one.                                            *
***********************************************************/
static struct hash_entry *HashTable_Insert(
    struct hash_table *pTable, /* must not be NULL */ 
    HASH hash,
    const char *key, /* must not be NULL */ 
    const char *value) /* must not be NULL */ 
{
  struct hash_entry **pEntry = HashTable_InternalLookup(pTable, hash, key);
  unsigned ValueLen;
  char *ValueCopy = NULL;

  REQUIRE(key != NULL);
  REQUIRE(READABLE(key));
  REQUIRE(value != NULL);
  REQUIRE(READABLE(value));

  ASSERT(READABLE(pEntry));

  if (*pEntry == NULL)
  {
    *pEntry = malloc(sizeof(**pEntry));
    if (*pEntry == NULL)
      return NULL;

    (*pEntry)->pChain = NULL;
    (*pEntry)->hash = hash;
    (*pEntry)->key = malloc((strlen(key) + 1) * sizeof(*(*pEntry)->key));
    (*pEntry)->value = NULL;
    
    if ((*pEntry)->key == NULL)
    {
      /* partial allocation failure - make it atomic */ 
      free(*pEntry);
      *pEntry = NULL;
      return NULL;
    }

    strcpy((*pEntry)->key, key);
  }

  ASSERT(WRITEABLE(*pEntry));
  
  /* lookup is case insensitive, replace to match new input case */ 
  /* currently there is no way to get the key back, so who cares
  strcpy((*pEntry)->key, key);
  */

  ValueLen = strlen(value);
  ValueCopy = malloc((ValueLen + 1) * sizeof(*ValueCopy));
  if (ValueCopy)
  {
    strcpy(ValueCopy, value);
    /* if free(NULL) crashes, your libraries are broken */ 
    free((*pEntry)->value);
    (*pEntry)->value = ValueCopy;
  }
  else if ((*pEntry)->value == NULL)
  {
    /* partial allocation failure - make it atomic */ 
    free((*pEntry)->key);
    free(*pEntry);
    *pEntry = NULL;
    return NULL;
  }

  ++pTable->nEntries;

  return *pEntry;
}

/***********************************************************
* struct hash_table *HashTable_Expand()                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct hash_table *pOld;                            *
*      The original hash table.  (must not be NULL)        *
*                         *  *  *  *                       *
* Returns a new hash table twice the size of the original  *
* or just returns the original if something fails.         *
***********************************************************/
static struct hash_table *HashTable_Expand(struct hash_table *pOld)
{
  const unsigned nOldSize = pOld->nTableSize;
  const unsigned nNewSize = nOldSize * 2;
  struct hash_table *pNew = HashTable_Alloc(nNewSize);
  struct hash_entry **pOrig, **pLow, **pHigh;
  struct hash_entry *pEntry, **pLowTail, **pHighTail;

  if (!pNew) return pOld;

  pNew->nTableSize = nNewSize;
  pNew->nEntries = pOld->nEntries;

  pOrig = pOld->Table + nOldSize;
  pLow  = pNew->Table + nOldSize;
  pHigh = pNew->Table + nNewSize;

  while (pOrig > pOld->Table)
  {
    --pOrig; --pLow; --pHigh;

    pEntry = *pOrig;
    pLowTail = pLow;
    pHighTail = pHigh;
    while (pEntry)
    {
      if (pEntry->hash & nOldSize)
      {
        (*pHighTail) = pEntry;
        pHighTail = &pEntry->pChain;
      }
      else
      {
        (*pLowTail) = pEntry;
        pLowTail = &pEntry->pChain;
      }
      pEntry = pEntry->pChain;
    }

    *pLowTail = NULL;
    *pHighTail = NULL;
  }

  free(pOld);
  return pNew;
}

/***********************************************************
* struct hash_table *HashTable_Destroy()                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct hash_table *pTable;                          *
*                         *  *  *  *                       *
* Frees the hash table and all of its entries.             *
***********************************************************/
struct hash_table *HashTable_Destroy(struct hash_table *pTable)
{
  struct hash_entry **pSlot, *pEntry, *pChain;

  if (pTable)
  {
    pSlot = pTable->Table + pTable->nTableSize;
    while (pTable->nEntries && pSlot > pTable->Table)
    {
      --pSlot;
      pEntry = *pSlot;
      while (pEntry)
      {
        pChain = pEntry->pChain;
        free(pEntry->key);
        free(pEntry->value);
        free(pEntry);
        pEntry = pChain;
      }
    }
    free(pTable);
  }

  return NULL;
}

/***********************************************************
* struct hash_table *HashTable_Set()                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - struct hash_table *pTable;                          *
*    - HASH hash;                                          *
*    - const char *key;                                    *
*    - const char *value;                                  *
*                         *  *  *  *                       *
* Associates a value with a key.                           *
* usage:  table = HashTable_Set(table, ...);               *
* The returned table may not be the same as the input, if  *
* the input table was NULL or needed to be enlarged.       *
***********************************************************/
struct hash_table *HashTable_Set(
    struct hash_table *pTable,
    HASH hash,
    const char *key,
    const char *value)
{
  unsigned nLimit;

  REQUIRE(key != NULL);
  REQUIRE(READABLE(key));
  REQUIRE(value != NULL);
  REQUIRE(READABLE(value));

  if (!pTable)
    /* arbitrary initial size */ 
    pTable = HashTable_Create(16);

  if (pTable)
  {
    /* aribtrary packing density: 3/4 */ 
    nLimit = pTable->nTableSize/2 + pTable->nTableSize/4;
    if (pTable->nEntries >= nLimit)
      pTable = HashTable_Expand(pTable);
    HashTable_Insert(pTable, hash, key, value);
  }
  return pTable;
}

/***********************************************************
* const char *HashTable_Get()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    -    const struct hash_table *pTable;                 *
*    -     HASH hash;                                      *
*    -     const char *key;                                *
*                         *  *  *  *                       *
* Retrieve the value associated with the given key, or NULL*
***********************************************************/
const char *HashTable_Get(
    const struct hash_table *pTable,
    HASH hash,
    const char *key)
{
  struct hash_entry *pEntry = NULL;

  if (pTable)
    pEntry = HashTable_Lookup((struct hash_table *)pTable, hash, key);
  return pEntry ? pEntry->value : NULL;
}

#ifndef NLOG
void HashTable_Dump(const struct hash_table *pTable)
{
  char sBuffer[4096];
  const char *sFormat;
  LPSTR values[8];
  unsigned nBuffer;
  const struct hash_entry * const *pEntry, *pChain;

  TRACE1("HashTable_Dump");
  if (!pTable)
  {
    ModuleLogWriteLine("no table", 8);
    return;
  }

  pEntry = pTable->Table + pTable->nTableSize;

  do
  {
    --pEntry;
    values[0] = ((LPSTR)pEntry - (LPSTR)pTable->Table);
    if (*pEntry == NULL)
      sFormat = "[%1!08X!]: NULL";
    else
    {
      sFormat = "[%1!08X!]: %2!08X! \"%3\" = \"%4\"";
      values[1] = (LPSTR)((*pEntry)->hash);
      values[2] = (LPSTR)((*pEntry)->key);
      values[3] = (LPSTR)((*pEntry)->value);
    }

    nBuffer = FormatMessage(
        FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
        sFormat,
        0,0,sBuffer,4096,values);

    ModuleLogWriteLine(sBuffer, nBuffer);

    if (*pEntry)
    {
      pChain = (*pEntry)->pChain;
      while (pChain)
      {
        sFormat = "         -> %2!08X! \"%3\" = \"%4\"";
        values[1] = (LPSTR)(pChain->hash);
        values[2] = (LPSTR)(pChain->key);
        values[3] = (LPSTR)(pChain->value);

        nBuffer = FormatMessage(
            FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
            sFormat,
            0,0,sBuffer,4096,values);

        ModuleLogWriteLine(sBuffer, nBuffer);

        pChain = pChain->pChain;
      }
    }
  }
  while (pEntry > pTable->Table);
}
#endif
